/*
 * A valueconstraint plugin for Netscape Directory Server.
 *
 * From the book "Understanding and Deploying LDAP Directory Services"
 *  by Timothy A. Howes, Mark C. Smith, and Gordon S. Good.
 */
#include <string.h>
#include <stdio.h>
#include "slapi-plugin.h"

/* Name and result code macros (to make the code easier to read) */
#define PLUGIN_NAME				"valueconstraint"
#define PLUGIN_RC_ERROR			(-1)
#define PLUGIN_RC_CONTINUE		0
#define PLUGIN_RC_RESULT_SENT	1

/* Macros to take care of platform specifics */
#ifdef _WIN32
#define PLUGIN_STRCASECMP			stricmp
#define PLUGIN_EXPORTED_FUNCTION	__declspec(dllexport)
#else
#define PLUGIN_STRCASECMP			strcasecmp
#define PLUGIN_EXPORTED_FUNCTION
#endif
 
/* Static variables to hold the plugin name and description. */
static Slapi_PluginDesc pdesc = { PLUGIN_NAME,
		"Howes/Smith/Good", "2.0", PLUGIN_NAME " plugin" };

/* Static variables to hold attribute constraint configuration. */
static const char *attr_to_check = "employeeType";
static const char *valid_values[] = {
	"Contractor", "Employee", "Intern", "Temp", NULL
};

/* Function prototypes */
static int valueconstraint_preadd( Slapi_PBlock *pb );
static int valueconstraint_premodify( Slapi_PBlock *pb );
static int valueconstraint_is_one_of( Slapi_PBlock *pb,
		const char *attrname, const char *val,
		const char *allowed[] );

/* Function valueconstraint_init(): initialization (called during
 *	server startup).
 */
PLUGIN_EXPORTED_FUNCTION int
valueconstraint_init( Slapi_PBlock *pb )
{
	int		rc = PLUGIN_RC_CONTINUE;

	slapi_log_error( SLAPI_LOG_PLUGIN, PLUGIN_NAME, "=> init\n" );

	/*
	 * Register the plugin version, plugin description,
	 * and two preoperation functions.
	 */
	if ( slapi_pblock_set( pb, SLAPI_PLUGIN_VERSION,
			SLAPI_PLUGIN_VERSION_01 ) != 0 ||
			slapi_pblock_set( pb, SLAPI_PLUGIN_DESCRIPTION,
			&pdesc ) != 0 ||
			slapi_pblock_set( pb, SLAPI_PLUGIN_PRE_ADD_FN,
			(void *)valueconstraint_preadd ) != 0 ||
			slapi_pblock_set( pb, SLAPI_PLUGIN_PRE_MODIFY_FN,
			(void *)valueconstraint_premodify ) != 0 ) {
		slapi_log_error( SLAPI_LOG_FATAL, PLUGIN_NAME,
				"init: a slapi_pblock_set() call failed\n" );
		rc = PLUGIN_RC_ERROR;
	}

	slapi_log_error( SLAPI_LOG_PLUGIN, PLUGIN_NAME,
			"<= init (%d)\n", rc );
	return rc;
}

/* Function valueconstraint_preadd(): called by the directory
 *	server before an add operation is processed.
 */
static int
valueconstraint_preadd( Slapi_PBlock *pb )
{
	int			rc = PLUGIN_RC_CONTINUE;
	Slapi_Entry	*entry = NULL;
	Slapi_Attr	*attr = NULL;

	slapi_log_error( SLAPI_LOG_PLUGIN, PLUGIN_NAME, "=> preadd\n" );

	/* Retrieve the entry that was sent in the Add operation */
	if ( slapi_pblock_get( pb, SLAPI_ADD_ENTRY, &entry ) != 0 ) {
		slapi_send_ldap_result( pb, LDAP_OPERATIONS_ERROR, NULL,
				PLUGIN_NAME ": unable to retrieve entry", 0, NULL );
		rc = PLUGIN_RC_RESULT_SENT;
	} else if ( slapi_entry_attr_find( entry, attr_to_check,
				&attr ) == 0 ) {
		/* Check the attribute values to make sure they are valid. */
		int			hint;
		Slapi_Value	*value;
		const char	*strvalue;

		for ( hint = slapi_attr_first_value( attr, &value );
				hint != -1;
				hint = slapi_attr_next_value( attr, hint, &value )) {
			strvalue = slapi_value_get_string( value );
			if ( !valueconstraint_is_one_of( pb, strvalue,
					attr_to_check, valid_values )) {
				rc = PLUGIN_RC_RESULT_SENT;
				break;
			}
		}
	}

	slapi_log_error( SLAPI_LOG_PLUGIN, PLUGIN_NAME,
			"<= preadd (%d)\n", rc );
	return rc;
}

/* Function valueconstraint_premodify(): called by the directory
 *	server before a modify operation is processed.
 */
static int
valueconstraint_premodify( Slapi_PBlock *pb )
{
	int			rc = PLUGIN_RC_CONTINUE;
	LDAPMod		*mp, **mods = NULL;
	int			i, j;

	slapi_log_error( SLAPI_LOG_PLUGIN, PLUGIN_NAME, "=> premodify\n" );

	/* Retrieve the modifications sent in the Modify operation */
	if ( slapi_pblock_get( pb, SLAPI_MODIFY_MODS, &mods ) != 0 ) {
		slapi_send_ldap_result( pb, LDAP_OPERATIONS_ERROR, NULL,
				PLUGIN_NAME ": unable to retrieve modifications",
				0, NULL );
		rc = PLUGIN_RC_RESULT_SENT;
	} else {
		/*
		 * Check the modify sub-operations to make sure the values
		 *	they contain are valid. Delete sub-operations are
		 *  ignored (any value may be deleted).
		 */
		for ( i = 0; rc == PLUGIN_RC_CONTINUE && mods[i] != NULL;
					++i ) {
			mp = mods[i];
			if ( 0 == slapi_attr_type_cmp( mp->mod_type,
					attr_to_check, SLAPI_TYPE_CMP_BASE )
					&& LDAP_MOD_DELETE !=
					( mp->mod_op & ~LDAP_MOD_BVALUES )) {
				for ( j = 0; mp->mod_bvalues[j] != NULL; ++j ) {
					if ( !valueconstraint_is_one_of( pb,
							mp->mod_bvalues[j]->bv_val,
							attr_to_check, valid_values )) {
						rc = PLUGIN_RC_RESULT_SENT;
						break;
					}
				}
			}
		}
	}

	slapi_log_error( SLAPI_LOG_PLUGIN, PLUGIN_NAME,
			"<= premodify (%d)\n", rc );
	return rc;
}

/* Function valueconstraint_is_one_of(): return a non-zero value
 *	if "val" is in the NULL-terminated "allowed" array and 0 if not.
 *	Comparisons are case-insensitive. If "val" is not present, a
 *	"constraint violation" error is sent to the LDAP client.
 */
static int
valueconstraint_is_one_of( Slapi_PBlock *pb, const char *val,
		const char *attrname, const char *allowed[] )
{
	int			i;
	char		*msg;
	const char	*fmt = "invalid value \"%s\" for %s";

	if ( val == NULL ) {
		return 1;	/* ignore NULL values */
	}

	for ( i = 0; allowed[i] != NULL; ++i ) {
		if ( PLUGIN_STRCASECMP( val, allowed[i] ) == 0 ) {
			return 1;	/* found it */
		}
	}
	
	/* not found: send back a constraint violation error */
	msg = slapi_ch_malloc( strlen(fmt) + strlen(val)
			+ strlen(attrname) + 1 );
	sprintf( msg, fmt, val, attrname );
	slapi_send_ldap_result( pb, LDAP_CONSTRAINT_VIOLATION,
			NULL, msg, 0, NULL );
	slapi_ch_free_string( &msg );
	return 0;	/* not found */
}
