/* -*- Mode: C; tab-width: 4; c-basic-offset: 4 -*- */
/*
 * A simple password resetting program for use by a helpdesk.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <ldap.h>
#ifdef _WINDOWS
#include <WINUSER.H>
#include <LMCONS.H>
#include "getopt.h"
#include "getpass.h"
#else
#include <unistd.h>
#include <pwd.h>
#endif


/*
 * Macros:
 */
#define DEFAULT_SEARCHBASE	"ou=people,dc=example,dc=com"
#define DEFAULT_LDAPHOST	"directory.example.com"
#define DEFAULT_LDAPPORT	LDAP_PORT
#define DEFAULT_WORDFILE	"/usr/dict/words"

#define LONGEST_WORD	100

#ifdef sun
#define GETPASS	getpassphrase	/* accepts a longer string */
#else
#define GETPASS	getpass			/* the old standby */
#endif

#ifdef _WINDOWS
#define SRANDOM( seed )	srand( seed )
#define RANDOM()		rand()
#else
#define SRANDOM( seed )	srandom( seed )
#define RANDOM()		random()
#endif


/*
 * Global variables:
 */
static int  quiet     = 0;	/* use the -q flag for quiet mode */
char		*wordfile = DEFAULT_WORDFILE;
char		*binddn = NULL;
char		*bindpwd = NULL;


/*
 * Function prototypes:
 */
static int resetpwd( LDAP *ld, const char *base,
		const char *userid, const char *newpwd );
static char *userid2dn( LDAP *ld, const char *base,
		const char *userid );
static void print_ldap_error( LDAP *ld, int lderr, const char *s );
static char *randompwd( void );
static char *randomword( void );
static int LDAP_CALL LDAP_CALLBACK get_rebind_credentials( LDAP *ld,
		char **dnp, char **passwdp, int *authmethodp,
		int freeit, void *arg );


/*
 * Functions:
 */
static void
usage( const char *progname )
{
	fprintf( stderr, "usage: %s [options] userid...\n", progname );
	fprintf( stderr, "where the options are:\n" );
	fprintf( stderr, "\t-H            help (display usage)\n" );
	fprintf( stderr, "\t-R            do not follow referrals\n" );
	fprintf( stderr, "\t-q            operate quietly\n" );
	fprintf( stderr, "\t-h host       LDAP server name\n" );
	fprintf( stderr, "\t-p port       LDAP server port\n" );
	fprintf( stderr, "\t-b basedn     base DN for searches\n" );
	fprintf( stderr, "\t-u bindid     user id to bind as\n" );
	fprintf( stderr, "\t-D binddn     DN to bind as\n" );
	fprintf( stderr, "\t-w passwd     bind password\n" );
	fprintf( stderr, "\t-n newpasswd  password to set\n" );
	fprintf( stderr, "\t-d file       words file, e.g. "
				DEFAULT_WORDFILE "\n" );
	exit( 2 );
}

int
main( int argc, char **argv )
{
    LDAP			*ld;
    int				i, c, errflg, rc;
	int				ldapv3 = LDAP_VERSION3;
	void			*follow_referrals = LDAP_OPT_ON;
	const char		*host = DEFAULT_LDAPHOST;
	int				port = DEFAULT_LDAPPORT;
	const char		*base = DEFAULT_SEARCHBASE;
	const char		*binduserid = NULL;
	const char		*newpwd = NULL;
	const char		*bindidstring = NULL;

#ifdef _WINDOWS
	char			useridbuf[ UNLEN + 1 ];
	DWORD			useridlen = sizeof(useridbuf);

	if ( GetUserName( useridbuf, &useridlen )) {
		binduserid = useridbuf;
	}
#else
	struct passwd	*pwdinfo;

	if ( NULL != ( pwdinfo = getpwuid( getuid() ))) {
		binduserid = strdup( pwdinfo->pw_name );
	}
#endif

	/*
	 * Process command line arguments.
	 */
	errflg = 0;
	while (( c = getopt( argc, argv, "HRqh:p:b:u:D:w:n:d:" )) != EOF ) {
		switch ( c ) {
		case 'H':	/* help */
			errflg = 1;
			break;
		case 'R':	/* do not follow referrals */
			follow_referrals = LDAP_OPT_OFF;
			break;
		case 'q':	/* quiet */
			quiet = 1;
			break;
		case 'h':	/* host */
			host = optarg;
			break;
		case 'p':	/* port */
			port = atoi( optarg );
			break;
		case 'b':	/* base DN */
			base = optarg;
			break;
		case 'u':	/* bind user id */
			binduserid = optarg;
			break;
		case 'D':	/* bind DN */
			binddn = optarg;
			break;
		case 'w':	/* bind password */
			bindpwd = optarg;
			break;
		case 'n':	/* new password */
			newpwd = optarg;
			break;
		case 'd':	/* words file */
			wordfile = optarg;
			break;
		default:
			errflg = 1;
		}
	}
	if ( errflg || optind >= argc ||
				( binduserid == NULL && binddn == NULL )) {
		usage( argv[ 0 ] );
	}

	if ( binddn != NULL ) {
		bindidstring = binddn;
	} else {
		bindidstring = binduserid;
	}

	/*
	 * Prompt for password if not provided on command line.
	 */
	if ( NULL == bindpwd || '\0' == *bindpwd ) {
		const char	*p = "LDAP password for";
		char		*prompt;

		prompt = malloc( strlen( p ) + strlen( bindidstring ) + 3 );
		if ( prompt == NULL ) {
			perror( "malloc" );
			return 1;
		}
		sprintf( prompt, "%s %s: ", p, bindidstring );
		bindpwd = GETPASS( prompt );
		free( prompt );
		if ( NULL == bindpwd || '\0' == *bindpwd ) {
			return 0;
		}
	}

    /*
     * Connect and bind to the directory server, looking up bind DN
	 * if a bind user id was provided.
     */
    if ( !quiet ) puts( "Contacting LDAP server..." );
	if (( ld = ldap_init( host, port )) == NULL ) {
		perror( host );
		return 1;
    }

	if ( ldap_set_option( ld, LDAP_OPT_PROTOCOL_VERSION,
				&ldapv3 ) != 0 ) {
		print_ldap_error( ld, ldap_get_lderrno( ld, NULL, NULL),
				"set protocol version" );
		ldap_unbind( ld );
		return 1;
	}

	if ( ldap_set_option( ld, LDAP_OPT_REFERRALS,
				follow_referrals ) != 0 ) {
		print_ldap_error( ld, ldap_get_lderrno( ld, NULL, NULL),
				"set referral following option" );
		ldap_unbind( ld );
		return 1;
	}

	if ( follow_referrals == LDAP_OPT_ON ) {
		if ( ldap_set_option( ld, LDAP_OPT_REBIND_FN,
					(void *)get_rebind_credentials ) != 0 ) {
			print_ldap_error( ld, ldap_get_lderrno( ld, NULL, NULL),
					"set referral rebind function" );
			ldap_unbind( ld );
			return 1;
		}
	}

	if ( binduserid != NULL ) {
		binddn = userid2dn( ld, base, binduserid );
		if ( binddn == NULL ) {
			return 1;
		}
	}

    rc = ldap_simple_bind_s( ld, binddn, bindpwd );
	if ( rc != LDAP_SUCCESS ) {
		print_ldap_error( ld, rc, bindidstring );
		ldap_unbind( ld );
		return 1;
    }


    /*
     * Process userid arguments, setting random passwords
     * for each one.
     */
    rc = 0;
    SRANDOM( time( NULL ));
    for ( i = optind; i < argc; ++i ) {
		if ( !quiet && i > optind ) {
			putchar( '\n' );
		}
		rc |= resetpwd( ld, base, argv[ i ], newpwd );
    }

    /*
     * close LDAP server connection and exit.
     */
    ldap_unbind( ld );
    return rc;
}


/*
 * resetpwd(): set an entry's password.
 * If newpwd is NULL, a random password is used.
 *
 * Returns 0 on success and 1 on failure.
 */
static int
resetpwd( LDAP *ld, const char *base,
		const char *userid, const char *newpwd )
{
    char	*dn, *pwd, *vals[2];
    int		rc;
    LDAPMod	mod, *mods[2];

    /*
     * Find the entry.
     */
    if ( !quiet ) printf( "Finding %s's entry...\n", userid );
	if (( dn = userid2dn( ld, base, userid )) == NULL ) {
		return 1;
    }

	if ( newpwd != NULL ) {
		pwd = (char *)newpwd;
	} else if (( pwd = randompwd()) == NULL ) {
		return 1;
	}

    /*
     * Replace the userPassword attribute.
     */
    if ( !quiet ) printf( "Modifying %s's entry...\n", userid );
    vals[0] = pwd;
    vals[1] = NULL;
    mod.mod_values = vals;
    mod.mod_type = "userPassword";
    mod.mod_op = LDAP_MOD_REPLACE;
    mods[0] = &mod;
    mods[1] = NULL;

    if (( rc = ldap_modify_s( ld, dn, mods )) == LDAP_SUCCESS ) {
		rc = 0;
		if ( quiet ) {
			printf( "%s's new password is: %s.\n", userid, pwd );
		} else {
			printf( "Password reset.  %s's new password is: %s.\n",
						userid, pwd );
			printf( "Don't forget to remind %s to reset "
				"his/her password right away!\n", userid );
		}
    } else {
		print_ldap_error( ld, rc, "modify" );
		rc = 1;
    }

	if ( newpwd == NULL ) {
		free( pwd );
	}
    ldap_memfree( dn );
    return rc;
}


/*
 * userid2dn(): Given an entry's userid, return its DN.
 *
 * The DN returned should be freed by passing it to
 * ldap_memfree(). If no entry is found, NULL is returned.
 */
static char *
userid2dn( LDAP *ld, const char *base, const char *userid )
{
    int			rc;
	const char	*filterpattern = "(&(objectClass=person)(uid=%s))";
    char		*filter, *dn = NULL;
    char		*attrs[] = { LDAP_NO_ATTRS, NULL };
    LDAPMessage	*e, *res = NULL;

    if (( filter = (char *)malloc( strlen( userid )
				+ strlen( filterpattern ) + 1 )) == NULL ) {
		perror( "malloc" );
		return NULL;
    }

    sprintf( filter, filterpattern, userid );
    rc = ldap_search_s( ld, base, LDAP_SCOPE_SUBTREE,
				filter, attrs, 0, &res );
    free( filter );

    if ( rc != LDAP_SUCCESS ) {
		print_ldap_error( ld, rc, "search" );
    } else if (( e = ldap_first_entry( ld, res )) == NULL ) {
		fprintf( stderr, "No match for user id %s.\n", userid );
    } else if ( ldap_next_entry( ld, e ) != NULL ) {
		fprintf( stderr, "%d matches for %s.\n",
					ldap_count_entries( ld, res ), userid );
    } else {
		dn = ldap_get_dn( ld, e );  /* found just one! */
    }

    ldap_msgfree( res );
    return dn;
}


/*
 * Callback function for LDAP bind credentials; used when
 * rebinding during referral and reference following.
 */
static int LDAP_CALL LDAP_CALLBACK
get_rebind_credentials( LDAP *ld, char **dnp, char **passwdp,
		int *authmethodp, int freeit, void *arg )
{
    if ( !freeit ) {
		*dnp = binddn;
		*passwdp = bindpwd;
		*authmethodp = LDAP_AUTH_SIMPLE;
    }

    return LDAP_SUCCESS;
}


/*
 * Display an LDAP error message.
 */
void
print_ldap_error( LDAP *ld, int lderr, const char *s )
{
    fprintf( stderr, "%s: %s\n", s, ldap_err2string( lderr ));
}


/*
 * randompwd(): generate a reasonably good password using some
 * random words from a file.  The password should be easy to
 * remember and able to be passed on verbally to a user.
 *
 * The algorithm we use is this:
 *  a) select two random words from the words file (w1 and w2).
 *  b) pick a random punctuation character (c).
 *  c) pick a random digit 0-9 (d)
 *  d) construct a candidate password as:  w1 c w2 d
 *  e) change the case of one of the letters at random
 */
static char *
randompwd( void )
{
    static char	*punctuation = "!@#$%&*-+=,.?";
    char	*pwd, *w1, *w2, *p;

    if (( w1 = randomword()) == NULL
				|| ( w2 = randomword()) == NULL ) {
		if ( w1 != NULL ) {
			free( w1 );
		}
		return NULL;
    }

    if (( pwd = (char *)malloc( LONGEST_WORD * 2 + 3 )) != NULL ) {
		sprintf( pwd, "%s%c%s%d", w1,
				punctuation[ RANDOM() % strlen( punctuation ) ], w2,
				RANDOM() % 10 );
		p = pwd + ( RANDOM() % strlen( pwd ));
		if ( isalpha( *p )) {
			if ( isupper( *p )) {
				*p = tolower( *p );
			} else {
				*p = toupper( *p );
			}
		}
	}

    free( w1 );
    free( w2 );
    return pwd;
}


/*
 * randomword(): return a random, lowercase word.
 *
 * Returns a malloc'd string if successful.
 * Returns NULL on error, e.g., if unable to access
 * the words file.
 *
 * This function assumes that all words in the file are
 * less than LONGEST_WORD characters long.
 */
static char *
randomword( void )
{
    static FILE	*fp = NULL;
    static long filesize = 0L;
    char	*word, *p;
    int		len;

    if ( fp == NULL ) {
		/*
		 * initialize: open the words file and
		 * determine its size
		 */
		if (( fp = fopen( wordfile, "r" )) == NULL
					|| fseek( fp, 0L, SEEK_END ) == -1
					|| ( filesize = ftell( fp )) == -1 ) {
			perror( wordfile );
			if ( fp != NULL ) {
				fclose( fp );
			}
			return NULL;
		}

		if ( filesize < LONGEST_WORD ) {
			fprintf( stderr, "%s: the words file must be at"
					" least %d bytes long.\n",
					wordfile, LONGEST_WORD );
			if ( fp != NULL ) {
				fclose( fp );
			}
			return NULL;
		}

		filesize -= LONGEST_WORD;
    }

    /*
     * seek to a random location in the file.
     */
    if ( fseek( fp, RANDOM() % filesize, SEEK_SET ) == -1 ) {
		perror( wordfile );
		return NULL;
    }

    if (( word = (char *)malloc( LONGEST_WORD )) == NULL ) {
		perror( "malloc" );
		return NULL;
    }

    /*
     * read the tail (remainder) of one word and then
     * the entire next word, which is what we will return.
     */
    fgets( word, LONGEST_WORD - 1, fp );
    if ( fgets( word, LONGEST_WORD - 1, fp ) == NULL ) {
		perror( wordfile );
		free( word );
		return NULL;
    }

    len = strlen( word ) - 1;
    if ( word[ len ] == '\n' ) {
		word[ len ] = '\0';
    }

    for ( p = word; *p != '\0'; ++p ) {
		if ( isupper( *p )) {
			*p = tolower( *p );
		}
    }

    return( word );
}
