/*
 * SimpleSite Servlet LDAP Application Example.
 *
 * From the 2nd Edition of the book:
 *   "Understanding and Deploying LDAP Directory Services"
 *   by Timothy A. Howes, Mark C. Smith, and Gordon S. Good.
 */
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import javax.naming.*;
import javax.naming.directory.*;

public class SimpleSiteServlet extends HttpServlet
{
/*
 * Data structures.
 */
private class fieldmap {
	boolean		mNewProfileOnly;
	String		mHtmlFieldName;
	int			mHtmlFieldSize;
	String		mLdapAttrName;
	String		mMapQuestFieldName;

	// fieldmap constructor
	private fieldmap( boolean newProfileOnly, String htmlFieldName,
				int htmlFieldSize, String ldapAttrName,
				String mapQuestFieldName ) {
		mNewProfileOnly = newProfileOnly;
		mHtmlFieldName = htmlFieldName;
		mHtmlFieldSize = htmlFieldSize;
		mLdapAttrName = ldapAttrName;
		mMapQuestFieldName = mapQuestFieldName;
	}
};

/*
 * Private data.
 */
private String	mApplName    = "SimpleSite";
private String	mMQURLPrefix = "http://www.mapquest.com/maps/map.adp?";
private String	mLdapURL;		// URL for the LDAP server
private String	mBaseDN;		// from "ldapBase" property
private String	mAddLocation;	// from "ldapAddLocation" property
private String	mAdder;			// from "ldapAdder" property
private String	mAdderPwd;		// from "ldapAdderPwd" property

private fieldmap[] mProfileFields = {
	new fieldmap( true,  "Email Address", 	50, "mail",         null ),
	new fieldmap( true,	 "Password",		30, "userPassword", null ),
	new fieldmap( false, "Name",			30, "cn",    		null ),
	new fieldmap( false, "Street Address",	40, "street",       "address" ),
	new fieldmap( false, "City",			40, "L",            "city" ),
	new fieldmap( false, "State",			 4,	"st",           "state" ),
	new fieldmap( false, "Zip Code",		12, "postalCode",	"zipcode" ),
	new fieldmap( false, "Country",			 4, "co",           "country" ),
	new fieldmap( false, "Areas of Expertise",
											50, "simpleExpertise",	null ),
	new fieldmap( false, "Preferred Email Format",
											 0, "simpleEmailFormat",null )
};

private String[] mLdapAttrsToRetrieve;

private String[] mObjectClassValues = {
	"inetOrgPerson",
	"simpleSiteObject"
};


/*
 * Public servlet methods.
 */
	/*
	 * init(): Perform essential initialization
	 */
	public void init(
			ServletConfig config )
			throws ServletException {
		super.init( config );

		// initialize our base DN, member RDN, and LDAP URL
		mBaseDN = getServletContext().getInitParameter( "ldapBase" );
		mAddLocation = getServletContext().getInitParameter(
				"ldapAddLocation" );
		mLdapURL = "ldap://"
				+ getServletContext().getInitParameter( "ldapServer")
				+ "/" + mBaseDN;

		// initialize the DN and password we use to create new entries
		mAdder = getServletContext().getInitParameter( "ldapAdder" );
		mAdderPwd = getServletContext().getInitParameter( "ldapAdderPwd" );

		// create a list of LDAP attributes we will retrieve
		mLdapAttrsToRetrieve = new String[ mProfileFields.length ];
		for ( int i = 0; i < mProfileFields.length; ++i ) {
			mLdapAttrsToRetrieve[i] = mProfileFields[i].mLdapAttrName;
		}
	}

	/*
	 * doGet(): handle an HTTP GET request
	 */
	public void doGet(
			HttpServletRequest request,
			HttpServletResponse response )
			throws IOException, ServletException {

		// prepare for response generation and dispatch to handler method
		response.setContentType( "text/html" );
		PrintWriter	writer = response.getWriter();

		String operation = request.getPathInfo();
		if ( operation == null ) {
			operation = "";
		} else {
			operation = operation.substring( 1 );
		}

		if ( operation.equals( "editprofile" )) {
			doEditProfile( request, response, writer );
		} else if ( operation.equals( "newprofile" )) {
			doNewProfile( request, response, writer );
		} else if ( operation.equals( "logout" )) {
			doLogout( request, response, writer );
		} else {
			unknownRequest( operation, "HTTP GET", writer );
		}
	}

	/*
	 * doPost(): handle an HTTP POST request
	 */
	public void doPost(
			HttpServletRequest request,
			HttpServletResponse response )
			throws IOException, ServletException {

		// prepare for response generation and dispatch to handler method
		response.setContentType( "text/html" );
		PrintWriter	writer = response.getWriter();

		String operation = request.getPathInfo();
		if ( operation == null ) {
			operation = "";
		} else {
			operation = operation.substring( 1 );
		}

		if ( operation.equals( "login" )) {
			doLogin( request, response, writer );
		} else if ( operation.equals( "find" )) {
			doFind( request, response, writer );
		} else if ( operation.equals( "saveprofile" )) {
			doSaveProfile( request, response, writer );
		} else {
			unknownRequest( operation, "HTTP POST", writer );
		}
	}

/*
 * Operation handlers.
 */
	/*
	 * doLogin(): handle a "login" HTTP POST sub-request
	 */
	private void doLogin(
			HttpServletRequest request,
			HttpServletResponse response,
			PrintWriter writer ) {

		// output the page header
		writePageHeader( mApplName + " - Login", writer );

		// retrieve the HTTP session
		HttpSession httpsession = request.getSession();

		// retrieve the login form variables
		String email = request.getParameter( "email" );
		if ( email == null || email.length() == 0 ) {
			reportError( "Login: please enter your email address",
					null, writer );
			return;
		}
		String pwd = request.getParameter( "pwd" );
		if ( pwd == null || pwd.length() == 0 ) {
			reportError( "Login: please enter your password", null, writer );
			return;
		}

		clearSessionValues( httpsession );

		try {
			DirContext ctx = createLDAPContext( null );

			// map the email address to a DN
			String partialDN = email2LDAPDN( ctx, email, writer );
			ctx.close();

			if ( partialDN != null ) {
				// try to authenticate
				ctx = createLDAPContext( partialDN, pwd );

				// store user information in the HTTP session
				httpsession.setAttribute( "id", email );
				httpsession.setAttribute( "partialDN", partialDN );
				httpsession.setAttribute( "pwd", pwd );

				// finish the page and close the LDAP connection
				writePageFooter( "<h3>Authentication succeeded.</h3>",
						true, writer  );
				ctx.close();
			}
		} catch ( Exception e ) {
			reportError( "Login:", e, writer );
		}
	}

	/*
	 * doLogout(): handle a "logout" HTTP GET sub-request
	 */
	private void doLogout(
			HttpServletRequest request,
			HttpServletResponse response,
			PrintWriter writer ) {

		// retrieve the HTTP session and clear all values
		HttpSession httpsession = request.getSession();
		clearSessionValues( httpsession );

		// since the session info. is gone, this will always redirect
		String id = getIDWithRedirect( httpsession, response, writer );
	}

	/*
	 * doFind(): handle a "find" HTTP POST sub-request
	 */
	private void doFind(	// HTTP POST method
			HttpServletRequest request,
			HttpServletResponse response,
			PrintWriter writer ) {

		// retrieve the HTTP session
		HttpSession httpsession = request.getSession();

		// generate start of page
		String id = getIDWithRedirect( httpsession, response, writer );
		if ( id == null ) return;
		writePageHeader( mApplName + " - Search Results", writer );

		// retrieve form variables and create an ANDed search filter.
		// For each value, we create an equality filter component,
		// except for "Areas of Expertise" where a substring
		// component is used if 3 chars. or more were entered.
		StringBuffer filter = new StringBuffer();
		Enumeration params = request.getParameterNames();
		while ( params.hasMoreElements()) {
			String paramName = (String)params.nextElement(); 
			String paramVal = request.getParameter( paramName );
			if ( paramVal != null && paramVal.length() > 0 ) {
				if ( paramName.equalsIgnoreCase( "simpleExpertise" )
						&& paramVal.length() >= 3 ) {
					filter.append( '(' + paramName + "=*" +
							escapedValue( paramVal ) + "*)" );
				} else {
					filter.append( '(' + paramName + '=' +
							escapedValue( paramVal ) + ')' );
				}
			}
		}
		if ( filter.length() == 0 ) {
			reportError( "Find: please provide some information",
					null, writer );
			return;
		}

		// restrict our search to person entries only
		filter.insert( 0, "(&(objectClass=person)" );
		filter.append( ')' );

		// search
		try {
			DirContext ctx = createLDAPContext( httpsession );

			SearchControls ctrls = new SearchControls(
					SearchControls.SUBTREE_SCOPE,
					10, 					// size limit
					1000 * 15,				// 15s time limit
					mLdapAttrsToRetrieve,	// attr list
					false,					// return entire object
					false					// dereference aliases
					);
			NamingEnumeration answer = ctx.search( "",
					filter.toString(), ctrls );

			int count = 0;
			SearchResult sr = null;
			while ( answer.hasMore()) {
				sr = (SearchResult)answer.next();
				displayOneEntry( sr.getAttributes(), writer );
				++count;
			}

			if ( count == 0 ) {
				reportError( "No matches", null, writer );
			} else if ( count > 1 ) {
				writer.println( "<h3>" + count + " people found.</h3>" );
			}

			writePageFooter( null, true, writer );

		} catch ( Exception e ) {
			reportError( "Find:", e, writer );
		}
	}

	/*
	 * doNewProfile(): handle a "newprofile" HTTP GET sub-request
	 */
	private void doNewProfile(
			HttpServletRequest request,
			HttpServletResponse response,
			PrintWriter writer ) {

		clearSessionValues( request.getSession());
		writePageHeader( mApplName + " - Create New Profile", writer );
		try {
			emitProfileForm( null, writer );
		} catch ( Exception e ) {
			reportError( "New profile:", e, writer );
		}
	}

	/*
	 * doEditProfile(): handle a "editprofile" HTTP GET sub-request
	 */
	private void doEditProfile(
			HttpServletRequest request,
			HttpServletResponse response,
			PrintWriter writer ) {

		// retrieve the HTTP session
		HttpSession httpsession = request.getSession();

		// retrieve the id plus DN and generate start of page
		String id = getIDWithRedirect( httpsession, response, writer );
		if ( id == null ) return;
		String partialDN = (String)httpsession.getAttribute( "partialDN" );
		writePageHeader( "Edit Profile for " + id, writer );

		try {
			// retrieve the user's profile information from the DS
			DirContext ctx = createLDAPContext( httpsession );

			Attributes attrs = ctx.getAttributes( partialDN,
					mLdapAttrsToRetrieve );
			emitProfileForm( attrs, writer );
			ctx.close();
		} catch ( Exception e ) {
			reportError( "Edit Profile:", e, writer );
		}
	}


	/*
	 * emitProfileForm(): common code for "editprofile" and "newprofile"
	 *     HTTP GET sub-requests.
	 */
	private void emitProfileForm(
			Attributes attrs,	// if null, create a "new profile" form
			PrintWriter writer )
			throws NamingException {

		boolean newProfile = ( attrs == null );

		writer.println( "<form action=\"saveprofile\" method=\"POST\">" );
		writer.println( "<table border=\"0\">\n<tbody>" );
		for ( int i = 0; i < mProfileFields.length; ++i ) {
			if ( !newProfile && mProfileFields[i].mNewProfileOnly ) {
				continue;
			}

			// retrieve the attribute value, if present
			Attribute attr = null;
			String val = "";

			if ( attrs != null ) {
				attr = attrs.get( mProfileFields[i].mLdapAttrName );
				if ( attr != null ) {
					val = (String)attr.get();
				}
			}

			if ( mProfileFields[i].mLdapAttrName.equalsIgnoreCase(
						"simpleEmailFormat" )) {
				// use radio buttons for preferred email format
				String	textChecked, htmlChecked;

				if ( val.length() == 0
							|| val.equalsIgnoreCase( "text/plain" )) {
					textChecked = " CHECKED";
					htmlChecked = "";
				} else {
					textChecked = "";
					htmlChecked = " CHECKED";
				}

				writer.println( "<tr><td>"
						+ mProfileFields[i].mHtmlFieldName
						+ ":</td>\n<td>" );
				writer.println( "<input type=\"radio\" name=\""
						+ mProfileFields[i].mLdapAttrName
						+ "\" value=\"text/plain\""
						+ textChecked + ">&nbsp;Text<br>" );
				writer.println( "<input type=\"radio\" name=\""
						+ mProfileFields[i].mLdapAttrName
						+ "\" value=\"text/html\""
						+ htmlChecked + ">&nbsp;HTML</td></tr>" );
			} else {
				String fieldType = "text";

				if ( mProfileFields[i].mLdapAttrName.equalsIgnoreCase(
						"userPassword" )) {
					// use an HTML password form field for userPassword
					fieldType = "password";
				}

				writer.println( "<tr><td>"
						+ mProfileFields[i].mHtmlFieldName
						+ ":</td>" );
				writer.print( "<td><input type=\"" + fieldType +
						"\" name=\"" +
						mProfileFields[i].mLdapAttrName + "\"" );
				if ( mProfileFields[i].mHtmlFieldSize > 0 ) {
					writer.print( " size=\""
							+ mProfileFields[i].mHtmlFieldSize
							+ "\"" );
				}
				writer.println( " value=\"" + val + "\"></td></tr>" );
			}
		}

		writer.println( "</tbody></table>" );
		if ( newProfile ) {
			writer.println( "<input type=\"hidden\"" +
					" name=\"_NewProfile\" value=\"TRUE\">" );
		}
		writer.println( "<input type=\"submit\" value=\"Save\">" );
		writer.println( "<input type=\"button\" value=\"Return to" +
			" Previous Page\" onClick=\"window.history.back()\">" );
		writer.println( "</form>" );
		writePageFooter( null, false, writer );
	}

	/*
	 * doSaveProfile(): handle a "saveprofile" HTTP POST sub-request
	 */
	private void doSaveProfile(	// HTTP POST method
			HttpServletRequest request,
			HttpServletResponse response,
			PrintWriter writer ) {

		// retrieve the HTTP session
		HttpSession httpsession = request.getSession();

		// determine whether we are processing a new profile
		boolean newProfile =
				( request.getParameter( "_NewProfile" ) != null );

		try {
			String id, partialDN, newPwd = null;
			DirContext ctx;

			if ( newProfile ) {
				// generate the id plus DN and generate start of page
				id = request.getParameter( "mail" );
				if ( id == null || id.length() == 0 ) {
					reportError( "New profile: email address required",
							null, writer );
					return;
				}
				partialDN = "mail=" + id + "," + mAddLocation;
				writePageHeader( "Creating Profile for " + id, writer );

				// establish an authenticated LDAP connection
				ctx = createLDAPContext( mAdder, mAdderPwd );
			} else {
				// retrieve the id plus DN and generate start of page
				id = getIDWithRedirect( httpsession, response, writer );
				if ( id == null ) return;
				partialDN = (String)httpsession.getAttribute( "partialDN" );
				writePageHeader( "Saving Profile for " + id, writer );

				ctx = createLDAPContext( httpsession );	// authenticate
			}

			// count the values
			int attrCount = 0;
			for ( int i = 0; i < mProfileFields.length; ++i ) {
				String ldapAttr = mProfileFields[i].mLdapAttrName;
				String val = request.getParameter( ldapAttr );

				if ( val != null ) {
					++attrCount;
					if ( ldapAttr.equalsIgnoreCase( "userPassword" )) {
						newPwd = val;
					}
				}
			}

			ModificationItem[] mods = null;
			BasicAttributes attrs = null;

			if ( newProfile ) {
				// create a collection of attributes
				attrs = new BasicAttributes();
				// include the objectClass values in the set
				BasicAttribute ocattr = new BasicAttribute( "objectClass" );
				for ( int i = 0; i < mObjectClassValues.length; ++i ) {
					ocattr.add( mObjectClassValues[i] );
				}
				attrs.put( ocattr );
			} else {
				// create an array of modification items
				mods = new ModificationItem[ attrCount ];
			}

			// populate the attrs collection or the mod item array
			attrCount = 0;
			for ( int i = 0; i < mProfileFields.length; ++i ) {
				String ldapAttr = mProfileFields[i].mLdapAttrName;
				String val = request.getParameter( ldapAttr );

				if ( val != null ) {
					BasicAttribute attr;

					if ( newProfile && val.length() > 0 ) {
						attr = new BasicAttribute( ldapAttr, val );
						attrs.put( attr );
						if ( ldapAttr.equalsIgnoreCase( "cn" )) {
							// construct a surname from the cn (simplistic)
							String snVal;
							int idx = val.indexOf( " " );
							if ( idx >= 0 ) {
								snVal = val.substring( idx + 1 );
							} else {
								snVal = val;
							}
							attrs.put( new BasicAttribute( "sn", snVal ));
						}
					} else if ( !newProfile ) {
						if ( val.length() > 0 ) {
							attr =  new BasicAttribute( ldapAttr, val );
						} else {	// remove all values
							attr =  new BasicAttribute( ldapAttr );
						}
						mods[ attrCount ] = new ModificationItem(
								ctx.REPLACE_ATTRIBUTE, attr );
					}
					++attrCount;
				}
			}

			// make the change via LDAP
			if ( newProfile ) {
				ctx.createSubcontext( partialDN, attrs );
				// store user information in the HTTP session
				httpsession.setAttribute( "id", id );
				httpsession.setAttribute( "partialDN", partialDN );
				httpsession.setAttribute( "pwd", newPwd );
			} else {
				ctx.modifyAttributes( partialDN, mods );
			}

			// finish the page and close the LDAP connection
			writePageFooter( "<h3>Profile successfully saved.</h3>",
					true, writer );
			ctx.close();
		} catch ( Exception e ) {
			reportError( "Edit Profile:", e, writer );
		}
	}

/*
 * LDAP utility methods.
 */
	/*
	 * createLDAPContext(): create an initial directory
	 *    context (open the LDAP connection). If partialDN is
	 *    not null, simple authentication is done.
	 */
	private DirContext createLDAPContext(
			String		partialDN,	// can be null
			String		pwd )
			throws NamingException {

		// initialize basic environment for JNDI's LDAP provider
		Hashtable env = new Hashtable();
		env.put( Context.INITIAL_CONTEXT_FACTORY,
				"com.sun.jndi.ldap.LdapCtxFactory" );
		env.put( Context.PROVIDER_URL, mLdapURL );
		env.put( Context.REFERRAL, "follow" );

		// handle optional simple bind
		if ( partialDN != null && partialDN.length() > 0 ) {
			String dn = partialDN + "," + mBaseDN;
			env.put( Context.SECURITY_AUTHENTICATION, "simple" );
			env.put( Context.SECURITY_PRINCIPAL, dn );
			env.put( Context.SECURITY_CREDENTIALS, pwd );
		}
		return new InitialDirContext( env );
	}

	/*
	 * createLDAPContext(): create an initial directory context based
	 *    on information found in an HTTP session.
	 */
	private DirContext createLDAPContext(
			HttpSession	httpsession )	// can be null
			throws NamingException {

		String dn = null, pwd = null;
		if ( httpsession != null ) {
			dn = (String)httpsession.getAttribute( "partialDN" );
			pwd = (String)httpsession.getAttribute( "pwd" );
		}
		return createLDAPContext( dn, pwd );
	}

	/*
	 * email2LDAPDN(): Find a person entry based on an email address.
	 *    Return a partial DN that is context-relative.
	 */
	private String email2LDAPDN(
			DirContext ctx,
			String email,
			PrintWriter writer ) {

		String partialDN = null;

		// construct the search filter and search controls
		String filter = "(&(objectClass=person)(mail=" +
				escapedValue( email ) + "))";
		String[] attrlist = { "1.1" };	// all we need is the DN
		SearchControls ctrls = new SearchControls(
				SearchControls.SUBTREE_SCOPE,
				100, 					// size limit
				1000 * 15,				// 15s time limit
				attrlist,				// attrs to retrieve
				false,					// return entire object
				false					// dereference aliases
				);

		// do the search
		try {
			NamingEnumeration answer = ctx.search( "", filter, ctrls );

			int count = 0;
			while ( answer.hasMore()) {
				SearchResult sr = (SearchResult)answer.next();
				if ( count == 0 ) {
					partialDN = sr.getName();
				}
				++count;
			}

			if ( count == 0 ) {
				reportError( "No matches for " + email, null, writer );
				partialDN = null;
			} else if ( count > 1 ) {
				reportError( count + " matches for " + email, null, writer );
				partialDN = null;
			}
		} catch ( Exception e ) {
			reportError( "Email lookup :", e, writer );
		}

		return partialDN;
	}

	/*
	 * escapedValue(): produce a copy of a string value, but with
	 *    special LDAP filter characters escaped as \HH (hex value).
	 */
	private String escapedValue( String value ) {
		StringBuffer escapedBuf = new StringBuffer();
		String specialChars = "*()\\";
		char c;

		for ( int i = 0; i < value.length(); ++i ) {
			c = value.charAt( i );
			if ( specialChars.indexOf( c ) >= 0 ) {	// escape it
				escapedBuf.append( '\\' );
				String hexString = Integer.toHexString( c );
				if ( hexString.length() < 2 ) {
					escapedBuf.append( '0' );
				}
				escapedBuf.append( hexString );
			} else {
				escapedBuf.append( c );
			}
		}
		return escapedBuf.toString();
	}

	/*
	 * Display one person's profile information.
	 */
	private void
	displayOneEntry( Attributes attrs, PrintWriter writer )
				throws NamingException {
		String mqURL = mMQURLPrefix;
		String email = null;

		writer.println( "<table border=\"0\"><tbody>" );
		for ( int i = 0; i < mProfileFields.length; ++i ) {
			Attribute attr = attrs.get(
					mProfileFields[i].mLdapAttrName );
			if ( attr == null ) {
				continue;
			}
			String val = (String)attr.get();
			if ( val != null && val.length() > 0 ) {
				writer.println( "<tr><td><b>" +
						mProfileFields[i].mHtmlFieldName +
						":</b></td>\n<td>" + val + "</td></tr>" );

				if ( mProfileFields[i].mMapQuestFieldName != null ) {
					mqURL = mqURL + "&" +
							mProfileFields[i].mMapQuestFieldName +
							"=" + val;
				}

				if ( mProfileFields[i].mLdapAttrName.equalsIgnoreCase(
							"mail" )) {
					email = val;
				}
			}
		}
		writer.println( "</tbody></table>\n" );

		if ( email != null ) {
			writer.println( "<a href=\"mailto:" + email +
					"\">Send Email</a>&nbsp;&nbsp;&nbsp" );
		}
		writer.println( "<a href=\"" + mqURL + "\" target=_new>"
			+ "Display Map</a><p>" );
	}


/*
 * General utility methods.
 */
	/*
	 * getIDWithRedirect(): Get the user's id from the HTTP session,
	 *     redirecting to the login page if not present.
	 */
	private String getIDWithRedirect(
			HttpSession httpsession,
			HttpServletResponse response,
			PrintWriter writer ) {

		String id = (String)httpsession.getAttribute( "id" );

		if ( id == null || id.length() == 0 ) {
			try {
				response.sendRedirect( "login.htm" );
			} catch ( Exception e ) {
				reportError( "Redirect to login:", e, writer );
			}
			id = null;
		}

		return id;
	}

	/*
	 * clearSessionValues(): clear old information that may be
	 *     in the HTTP session.
	 */
	private void clearSessionValues( HttpSession httpsession ) {
		httpsession.setAttribute( "id", "" );
		httpsession.setAttribute( "partialDN", "" );
		httpsession.setAttribute( "pwd", "" );
	}

	/*
	 * writePageHeader(): Output an HTML page header with title.
	 */
	private void writePageHeader(
			String title,
			PrintWriter writer ) {

		if ( title == null ) {
			title = mApplName;
		}
		writer.println( "<html>" );
		writer.println( "<head><title>" + title + "</title></head>" );
		writer.println( "<body>\n<center><h2>" + title + "</h2></center>" );
	}

	/*
	 * writePageFooter(): Output an HTML page footer
	 */
	private void writePageFooter(
			String statusMessage,
			boolean displayActions,
			PrintWriter	writer ) {

		if ( statusMessage != null ) {
			writer.println( "<p>" + statusMessage );
		}
		if ( displayActions ) {
			writer.println( "<form>" );
			writeHREFButton( "Find Person", "../find.htm", writer );
			writeHREFButton( "Edit Profile", "editprofile", writer );
			writeHREFButton( "Log Out ", "logout", writer );
			writer.println( "</form>" );
		}
		
		writer.println( "</body>\n</html>" );
	}

	/*
	 * writeHREFButton(): Output a button that is a link.
	 */
	private void writeHREFButton(
			String label,
			String url,
			PrintWriter writer ) {

		writer.println( "<input type=\"button\" value=\""
				+ label + "\" onClick=\"window.location.href='"
				+ url + "'\">&nbsp;&nbsp;" );
	}


	/*
	 * unknownRequest(): Output an "unknownRequest" HTML error
	 *    page (entire page).
	 */
	private void unknownRequest(
			String msg,
			String type, // get, post, ...
			PrintWriter writer ) {

		writePageHeader( mApplName + " - unknown request", writer );
		reportError( "Unknown " + type + " request: (" + msg + ")",
				null, writer );
	}

	/*
	 * reportError(): Output an HTML error page (all but page header).
	 */
	private void reportError(
			String msg,
			Exception e,
			PrintWriter writer ) {

		writer.println( "<p><b>" + msg + "</b><br>" );
		if ( e != null ) {
			writer.println( "Error: " + e + "<br>" );
		}
		writer.println( "<form><input type=\"button\" "
				+ "value=\"Try Again\" "
				+ "onClick=\"window.history.back()\"></form>" );
		writePageFooter( null, false, writer );
	}
}
