#!/usr/local/bin/perl
#
# ldapsync -- Perl 5 script that synchronizes a
#	comma-separated text file of attribute values.
#
# From the 2nd Edition of the book:
#	"Understanding and Deploying LDAP Directory Services"
#	by Timothy A. Howes, Mark C. Smith, and Gordon S. Good.
#
# usage: ldapsync.pl < file
#
# Where the contents of file are of the form:
#	joinAttrName,attrname1,attrname2...
#	joinAttrValue,attrvalue1,attrvalue2...
#	joinAttrValue,attrvalue1,attrvalue2...
#	...
#
# For example:
#	uid,cn,telephoneNumber
#	bjensen,Barbara Jensen,+1 650 555-1212
#	cms,Christina Smith
#	jjones,John Jones,+1 734 555-1212
#	ajackson,,+1 810 555-1212
#
# Requires: PerLDAP
#

use Mozilla::LDAP::Conn;

# LDAP server information:
$ldapSearchBase = "dc=example,dc=com";
$ldapHost       = "ldap.example.com";
$ldapPort       = "389";
$ldapBindDN     = "cn=LDAP Sync Tool,ou=Special Users,"
							.  $ldapSearchBase;
$ldapBindPW     = "Tb58-#Wxza";

# Start of main: 

logInfo( (scalar localtime) . " - ldapsync started" );

# Open an authenticated connection to the LDAP server.
$ldap = new Mozilla::LDAP::Conn( $ldapHost, $ldapPort,
		$ldapBindDN, $ldapBindPW );
if ( ! $ldap ) {
	logError( "Unable to connect to server at "
			. "ldap://$ldapHost:$ldapPort" );
	exit 1;
}

# For each line of input, search for the directory entry
# corresponding to the first field, and see if its value
# for the second field needs to be updated
while ( <STDIN> ) {
# Read one line.
	$line = $_;

# skip lines that begin with '#' (comments)
	if ( $line =~ /^#/ ) {
		next;
	}

# discard newline and return characters; break at commas
	chop $line;
	if ( $line =~ /\r$/ ) {
		chop $line;
	}
	@args = split( /,/, $line );

# If this is the first line, read attribute names
	if ( ! $joinAttrName ) {
		$joinAttrName = @args[0];
		@attrNameList = @args;
		next;
	}

	++$changeCount;

# Parse join attribute and attribute values to update
	$joinAttrValue = @args[0]; 
	@attrValueList = @args;

# Search for entry with uid equal to the join attribute
    $filter = "(&(objectClass=person)"
			. "($joinAttrName=$joinAttrValue))";
    $entry = $ldap->search( $ldapSearchBase, "subtree", 
				$filter, 0, @attrNameList );

# Found a match - update if necessary
	$matchLogString = "with $joinAttrName $joinAttrValue";
	$entryCount = 0;
	for ( $tmpEntry = $entry; $tmpEntry;
				$tmpEntry = $ldap->nextEntry ) {
		++$entryCount;
	}
	if ( $entryCount == 0 ) {
		logError( "Found no entry $matchLogString." );
		++$notFoundCount;
	} elsif ( $entryCount > 1 ) {
		logError( "$entryCount entries found $matchLogString." );
		++$multipleMatchCount;
	} else {
		logInfo( "Found one entry $matchLogString."
				. " Checking for updates..." );
		$entryNeedsUpdating = 0;
		for ( $i = 1; $i < @attrNameList.size; ++$i ) {
			if ( $i > @attrValueList.size ) {
# no more new values: break out of for loop.
				done;
			}
			if ( @attrValueList[$i] eq "" ) {
# new value is empty: try the next one
				next;
			}

# Replace attribute's values if new value does not match old
			if ( ! $entry->hasValue( @attrNameList[$i],
					@attrValueList[$i], 0 )) {
				$entry->setValues( @attrNameList[$i],
						@attrValueList[$i] );
				logInfo( "    - replace @attrNameList[$i]"
						. " with @attrValueList[$i]" ); 
				$entryNeedsUpdating = 1;
			}
		}

# Update entry over LDAP if needed
		if ( $entryNeedsUpdating ) {
			logInfo( "Modifying entry $matchLogString" );
			$ldap->update( $entry );
			if ( $ldap->getErrorCode()) {
				logError( "Modify failed for entry"
						. " $matchLogString: "
						. $ldap->getErrorString() );
				++$updateFailureCount;
			} else {
				logInfo( "Modify done." );
				++$updateSuccessCount;
			}
		}
	}
}

# Close LDAP connection and clean up
$ldap->close;

# Print an activity summary and determine exitCode
logInfo( "--------------------------------------" );
logInfo( "Summary:" );
$exitCode = 0;
if ( $changeCount > 0 ) {
	logInfo( "  $changeCount change records were processed." );
	if ( $updateSuccessCount > 0 ) {
		logInfo( "  $updateSuccessCount entries were updated." );
	} else {
		logInfo( "  No entries were updated." );
	}
	if ( $updateFailureCount > 0 ) {
		logInfo( "  $updateFailureCount failed update(s)." );
		$exitCode = 1;
	}
	if ( $notFoundCount > 0 ) {
		logInfo( "  $notFoundCount entries were not found." );
		$exitCode = 1;
	}
	if ( $multipleMatchCount > 0 ) {
		logInfo( "  $multipleMatchCount $joinAttrName values"
				. " matched more than one entry." );
		$exitCode = 1;
	}
}

# That's all folks!
logInfo( (scalar localtime) . " - ldapsync finished." );
exit $exitCode;
# End of main.


# Start of logError():
sub
logError {
	local( $msg ) = @_;

	print "ERROR: $msg\n";
}
# End of logError().


# Start of logInfo():
sub
logInfo {
	local( $msg ) = @_;

	print "INFO.: $msg\n";
}
# End of logInfo().
